#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "bklist.h"

/***********************************************************************
  The file format is as follows:

   file: record*

   record:  record_size 4 bytes
            type        1 byte
            date, title, url string

   if type is 'E', then the parser continues with the parent node.
 ***********************************************************************/

static int read_force(FILE *f, unsigned char *buf, int size)
{
  int R;
  while(size)
  {
     R= fread(buf, 1, size, f); 
     if (R<=0) return 1;
     buf+= R;
     size-= R;
  }
  return 0;
}

static int read_size(FILE *f, int *sz)
{
  unsigned char buf[4];
  if (read_force(f, buf,4)) return 1;
  *sz= (buf[3]<<24) | (buf[2]<<16) || (buf[1]<<8) | buf[0];
  if (*sz<0 || *sz>4096) return 1;
  return 0;
}

static char *getstr(unsigned char *start, unsigned char *end, int count)
{
  unsigned char *nulls[10];
  int nend;
  unsigned char *k;
  nend= 0;

  for(k=start;nend<10 && k!=end;k++)
   if (!k[0]) nulls[nend++]= k;

  if (count>nend) return strdup("");
  if (count==1) return strdup((char*) start);
  return strdup((char*) nulls[count-2]+1);
}


  // The top pointer is used only for its children.
  // It will of type 'F'.
int read_bkelts(char *fn, bkelt_t **Rtop)
{
  FILE *f;
  int size;
  unsigned char *buf; int bs;
  bkelt_t *top,*parent,*elt;
  bkelt_t *stack[200]; int tos;

  tos= -1;

  f= fopen(fn,"rb");
  if (!f) return 1;
  bs= 1000;
  buf= malloc(bs);

  parent= top= bkelt('F',NULL,NULL,NULL);

  while(1)
  {
    if (read_size(f, &size)) break;
    if (size>bs) { bs= size; buf= realloc(buf, bs); }
    if (read_force(f, buf, size)) break;
    if (buf[0]=='E')
    { 
       if (tos==-1) continue;
       parent= stack[tos];
       tos--;
       continue;
    }
    elt= bkelt(buf[0],
                 getstr(buf+1, buf+size, 3),     // url
                 getstr(buf+1, buf+size, 2),    // title
                 getstr(buf+1, buf+size, 1));   // date
    bkelt_add_child(parent, elt);
    if (elt->type=='F')
    {
      tos++;
      stack[tos]= parent;
      parent= elt;
    }
  }
  fclose(f);
  *Rtop= top;
  return 0;
}

static int write_force(FILE *f, unsigned char *buf, int len)
{
  int R;
  while(len)  
  {
    R= fwrite(buf,1,len,f);
    if (R<=0) return 1;
    buf+= R;
    len-= R;
  }
  return 0;
}

static int write_elt(FILE *f, bkelt_t *B)
{
  static unsigned char *buf;
  static int bufsiz;
  bkelt_t *P;
  int O;
  int L1,L2,L3;
  int size;

  L1= strlen(B->date)+1;
  L2= strlen(B->title)+1;
  L3= strlen(B->url)+1;
  size=4+ 1+ L1 + L2 + L3;  
  if (size>bufsiz)
  {
     bufsiz= size;
     if (bufsiz<10) bufsiz= 10;
     if (!buf) buf= malloc(bufsiz);
         else  buf= realloc(buf, bufsiz);
  }
  buf[0]= size;
  buf[1]= size>>8;
  buf[2]= size>>16;
  buf[3]= size>>24;
  buf[4]= B->type;
  O= 5;
  memcpy(buf+O, B->date, L1);
  O+= L1;
  memcpy(buf+O, B->title, L2);
  O+= L2;
  memcpy(buf+O, B->url, L3);

  if (write_force(f, buf, size)) return 1;
  if (B->type!='F') return 0;
  for(P=B->children->next;P!=B->children;P=B->next)
    if (write_elt(f, P)) return 1;
     // size= 4
  buf[0]= 4; buf[1]= 0; buf[2]= 0; buf[3]= 0;
     // type= E
  buf[4]= 'E';
     // three nulls
  buf[5]= 0; buf[6]= 0; buf[7]= 0;
  if (write_force(f, buf, 8)) return 1;
  return 0;
}


int write_bkelts(char *fn, bkelt_t *top)
{
  bkelt_t *P;
  FILE *f;
  int R;
  
  f= fopen(fn,"wb");
  if (!f) return 1;
  R= 0;
  for(P=top->children->next;P!=top->children;P=P->next)
    if (write_elt(f, P)) { R= 1; break; }
  fclose(f);
  return R;
}

bkelt_t *bkelt(int type,char *url,char *title,char *date)
{
  bkelt_t *B;
  B= calloc(1,sizeof(*B));
  B->type= type;
  B->url= url;
  B->title= title;
  B->date= date;

  if (type=='F')
  {
    B->children= calloc(1,sizeof(*B));
    B->children->prev= B->children->next= B->children;
  }
  B->next= B->prev= B;
  return B;
}

void bkelt_add_child(bkelt_t *parent, bkelt_t *node)
{
  if (parent->type!='F') return ;
  node->prev= parent->children->prev;
  node->next= parent->children;
  parent->children->prev->next= node;
  parent->children->prev= node;
  node->parent= parent;
}
