#include <stdio.h>
#include <unistd.h>
#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include "bklist.h"
#include "gfx.h"
#include "data.c"
#include "utf8.h"

int html_to_bklist(char *fn, bkelt_t **Rtop);

static char *strcom2(char *A,char *B)
{
  char *R;
  R= malloc(strlen(A)+strlen(B)+1);
  strcpy(R,A);
  strcat(R,B);
  return R;
}

static char* cfg_file_name(char *base)
{
  char *cfg_home;
  cfg_home= getenv("XDG_CONFIG_HOME");
  if (!cfg_home)
  {
    char *home;
    home= getenv("HOME");
    if (!home) home= "/tmp";
    cfg_home= strcom2(home,"/.config");
  }
  return strcom2(strcom2(cfg_home,"/bkedit/"),base);
}

static int read_config_file(char *base, char **R)
{
  unsigned char *d; size_t l;
  if (file_read_all(cfg_file_name(base),&d, &l)) return 1;
  *R= (char*) d;
  return 0;
}


static int spc;
void print_children(bkelt_t *node);

static void print_spc() { int i; for(i=0;i<spc;i++) printf(" "); }

void print_elt(bkelt_t *node)
{
  print_spc();
  printf("[%c][url=%s][title=%s][date=%s]\n", node->type,
         node->url, node->title, node->date);
  if (node->type=='F')
    print_children(node);
}

void print_children(bkelt_t *node)
{
  bkelt_t *P;
  spc+= 3;
  for(P=node->children->next;P!=node->children;P=P->next)
    print_elt(P);
  spc-= 3;
}

void print_bklist(bkelt_t *top)
{
  print_children(top);
}

typedef struct
{
  int fn, list, url, title;
  char *file_name;
  bkelt_t *topnode;
  int selected;
  bkelt_t *selitem;
  bkelt_t *folder;
} edi_t;

static edi_t edi[]= 
  {
    { Wfnleft, Wleft, Wurlleft, Wtitleleft },
    { Wfnright, Wright, Wurlright, Wtitleright },
  };

static void clear_fields(int idx)
{
  gfx_set_text(edi[idx].title,"");
  gfx_set_text(edi[idx].url,"");
  edi[idx].selected= -1;
  edi[idx].selitem= NULL;
}


static void list_insert(int idx, int pos, bkelt_t *P)
{
  gfx_insert(edi[idx].list, pos, P->title,P, 
                    P->type=='F' ? GFXRES_directory : NULL);
}

static void list_append(int idx, bkelt_t *P)
{
  list_insert(idx, -1, P);
}

static void display_folder(int idx, bkelt_t *folder)
{
  bkelt_t *P;
  gfx_clear(edi[idx].list);
  for(P=folder->children->next;P!=folder->children;P=P->next)
    list_append(idx, P);
  clear_fields(idx);
  edi[idx].folder= folder;
}

static void select_item(int idx, int key, bkelt_t *elt,int gsel)
{
  if (!elt || key==-1) return ;
  gfx_set_text(edi[idx].url, elt->url);
  gfx_set_text(edi[idx].title, elt->title);
  edi[idx].selected= key;
  edi[idx].selitem= elt;
  if (gsel)
    gfx_set_selected(edi[idx].list, key);
}

static void load(int idx, char *fn)
{
  bkelt_t *top;
  if (html_to_bklist(fn, &top)) return ;
  edi[idx].topnode= top;
  gfx_clear(edi[idx].list);
  gfx_set_text(edi[idx].fn, fn);
  edi[idx].file_name= fn;
  display_folder(idx, edi[idx].topnode);
}

static void load_file(int idx)
{
  char *fn;
  static char *html_files[]= 
  { "HTML (.html .htm)", NULL };
  fn= gfx_select_file(0, "", html_files);
  if (!fn) return ;
  load(idx,fn);
}

static void double_click(int idx, bkelt_t *elt)
{
  char *prg;
  char *cmd;
  char *argv[3];
  if (!elt) return;
  if (elt->type=='F')
  { 
    if (elt->children) 
       display_folder(idx, elt);
    return ;
  }
  if (!elt->url || !elt->url[0]) return ;
  prg= cfg_file_name("browse.sh");
  if (access(prg,R_OK|X_OK))
    prg="firefox";

  argv[0]=prg;
  argv[1]=elt->url;
  argv[2]=NULL;

  if (fork()) return ;
  execvp(argv[0],argv);
  exit(0); 
}

static bkelt_t *delete_item(int idx)
{
  bkelt_t *P;
  if (edi[idx].selitem==NULL || edi[idx].selected==-1) return NULL;
  gfx_remove(edi[idx].list, edi[idx].selected, NULL);

  P= edi[idx].selitem;
  clear_fields(idx);
  P->next->prev= P->prev;
  P->prev->next= P->next;
  return P;
}

static void move_item(int src,int dst)
{
  bkelt_t *P;
  if (!edi[dst].folder || edi[dst].folder->type!='F' ||
         !edi[dst].folder->children) return ;
  P= delete_item(src);
  if (!P) return ;
  list_append(dst, P);
  bkelt_add_child(edi[dst].folder, P);
}

static void parent_folder(int idx)
{
  bkelt_t *P;
  P= edi[idx].folder;
  if (!P || !P->parent) return ;
  display_folder(idx, P->parent);
}

static int selected(int idx, int *key, bkelt_t **elt)
{
  if (edi[idx].selected==-1) return 0;
  if (edi[idx].selitem==NULL) return 0;
  *key= edi[idx].selected;
  *elt= edi[idx].selitem;
  return 1;
}

static void move_item_up(int idx)
{
  int key;
  bkelt_t *elt;
  edi_t *E= edi+idx;
  bkelt_t *FC,*anchor;
  if (!selected(idx,&key, &elt)) return ;
  if (!E->folder || !(FC=E->folder->children)) return ;
  anchor= elt->prev;
  if (anchor==FC) return ;
  delete_item(idx); 

  elt->prev= anchor->prev;
  elt->next= anchor;
  anchor->prev->next= elt;
  anchor->prev= elt;

  list_insert(idx, key-1, elt);
  select_item(idx, key-1,elt,1);
}

static void move_item_down(int idx)
{
  int key;
  bkelt_t *elt;
  edi_t *E= edi+idx;
  bkelt_t *FC,*anchor;
  if (!selected(idx,&key, &elt)) return ;
  if (!E->folder || !(FC=E->folder->children)) return ;
  anchor= elt->next;
  if (anchor==FC) return ;
  delete_item(idx); 

  elt->next= anchor->next;
  elt->prev= anchor;
  anchor->next->prev= elt;
  anchor->next= elt;

  list_insert(idx, key+1, elt);
  select_item(idx, key+1,elt,1);
}

static void update_title(int idx)
{
  int key;
  bkelt_t *elt;

  if (!selected(idx, &key,&elt)) return ;
  gfx_remove(edi[idx].list, key, NULL);
  elt->title= gfx_get_text(edi[idx].title);
  list_insert(idx, key, elt);
  select_item(idx, key, elt,1);
}

static int all_spaces(char *txt)
{
  while(*txt) if (*txt>32) return 0; else txt++;
  return 1;
}

static void make_folder(int idx)
{
  gfx_event_t event;
  char *title;
  bkelt_t *P;
  if (!edi[idx].folder || !edi[idx].folder->children) return ;
  gfx_SCRmkfolder();

  while(1)
   if (gfx_event(NULL,&event))
   switch(event.type)
   {
   case GFXEV_QUIT:
   case WfolCANCEL:
     gfx_leave_screen();
     return ;
   case WfolOK:
     title= gfx_get_text(WfolNAME);
     gfx_leave_screen();
     if (!title || !title[0] || all_spaces(title)) return ;
     P=bkelt('F', "",title, "");
     bkelt_add_child(edi[idx].folder, P);
     list_append(idx, P);
     return ;
   }
}
static void write_children(FILE *out, bkelt_t *folder);

static int is_special(uint32_t chr)
{
  if (chr>=127) return 1;
  if (strchr("\"&<>'",chr)) return 1;
  return 0;
}

static void hout(FILE* out, char *z)
{
  static uint32_t *wide;
  static size_t wsiz;
  size_t i;
  size_t zl, wlen;
  zl= strlen(z);
  if (!wide || zl>wsiz)
  {
     size_t asiz;
     wsiz= zl + zl/2 + 5;
     asiz= wsiz * sizeof(wide[0]);
     if (wide) wide= realloc(wide, asiz);
          else wide= malloc(asiz);
  }
  utf8_decode_string((uint8_t*) z,strlen(z), wide, &wlen);
  for(i=0;i<wlen;i++)
    if (is_special(wide[i]))
      fprintf(out,"&#%d;",wide[i]);
    else 
      fprintf(out,"%c", wide[i]);
}

static void write_elt(FILE *out, bkelt_t *E)
{
  fprintf(out,"<DT>");
  if (E->type=='F') 
  {
    fprintf(out,"<H3 ADD_DATE=\"%s\">",E->date);
    hout(out, E->title);
    fprintf(out,"</H3>\n");
    write_children(out, E);
  }
  else
  {
    fprintf(out,"<a href=\""); 
    hout(out, E->url); 
    fprintf(out,"\" ADD_DATE=\"%s\">", E->date);
    hout(out, E->title);
    fprintf(out,"</A>\n");
  }
}

static void write_children(FILE *out, bkelt_t *folder)
{
  bkelt_t *P;
  if (!folder || !folder->children) return ;
  fputs("<DL><p>",out);
  for(P=folder->children->next;P!=folder->children;P=P->next)
    write_elt(out, P);
  fputs("</DL>",out);
}

static void save_file(int idx)
{
  edi_t *E;
  char *hdr, *ftr;
  FILE *out;
  E= edi + idx;
  if (!E->file_name || !E->topnode) return ;

  if (read_config_file("hdr.txt", &hdr) ||
      read_config_file("ftr.txt", &ftr))
   { hdr= def_hdr; ftr= def_ftr; }

  out= fopen(E->file_name, "wb");
  if (!out) return ;
  fputs(hdr, out);
  write_children(out,E->topnode);
  fputs(ftr,out);
  fclose(out);
  if (hdr!=def_hdr) free(hdr);
  if (ftr!=def_ftr) free(ftr);
}

int main(int argc,char **argv)
{
#if 0
  bkelt_t *top;

  if (html_to_bklist(argv[1], &top)) return 1;
  print_bklist(top);
  return 0;
#endif

  edi[0].selected= edi[1].selected= -1;

  gfx_init(NULL,900, 800, "Bookmarks Editor");

  gfx_SCREEN_0();

  if (argc>1) load(0,argv[1]);
  if (argc>2) load(1,argv[2]);
  while(1)
  {
    gfx_event_t event;
    if (!gfx_event(NULL,&event)) continue;
    switch(event.type)
    {
    case GFXEV_QUIT: goto done;
    case Wloadleft: load_file(0); break;
    case Wloadright: load_file(1); break;
    case GFXEV(Wleft, DOUBLE_CLICK): double_click(0, event.usr); break;
    case GFXEV(Wright, DOUBLE_CLICK): double_click(1, event.usr); break;
    case Wleft:  select_item(0, event.key, event.usr,0); break;
    case Wright: select_item(1, event.key, event.usr,0); break;
    case Wdelleft:  delete_item(0); break;
    case Wdelright: delete_item(1); break;
    case Wputright: move_item(0, 1); break;
    case Wputleft:  move_item(1,0); break;
    case Wparentleft:  parent_folder(0); break;
    case Wparentright: parent_folder(1); break;
    case Wupleft:  move_item_up(0); break;
    case Wupright: move_item_up(1); break;
    case Wdnleft:  move_item_down(0); break;
    case Wdnright: move_item_down(1); break;
    case Wupdateleft: update_title(0); break;
    case Wupdateright: update_title(1); break;
    case Wmkdirleft: make_folder(0); break;
    case Wmkdirright: make_folder(1); break;
    case Wsaveleft: save_file(0); break;
    case Wsaveright: save_file(1); break;
    }
  }

 done:
  return 0;
}


