#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "gumbo.h"
#include "fread.h"
#include "bklist.h"
#include "buffer.h"

static int print_spc(int X)
{
  while(X--) printf(" ");
}

static void prt_all(GumboNode *node)
{
  GumboElement *elt;
  GumboText *text;
  int i;
  static int spc;
  if (node->type!=GUMBO_NODE_WHITESPACE)
  print_spc(spc);
  switch(node->type)
  {
  case GUMBO_NODE_DOCUMENT: printf("document\n"); break;
  case GUMBO_NODE_ELEMENT:
    elt= &node->v.element;
    printf("(%s)\n", gumbo_normalized_tagname(elt->tag));
    spc+=2;
    for(i=0;i<elt->attributes.length;i++)
    {
      GumboAttribute *A;
      print_spc(spc);
      A= elt->attributes.data[i];
      printf("%s=%s\n", A->name, A->value);
    }
    spc+= 3;
    for(i=0;i<elt->children.length;i++)
     prt_all(elt->children.data[i]);
    spc-= 5;
    break;
  case GUMBO_NODE_TEXT:
  case GUMBO_NODE_CDATA:
  case GUMBO_NODE_COMMENT:
    text= &node->v.text;
    printf("T[%s]\n",text->text);
    break;
  case GUMBO_NODE_TEMPLATE: printf("template\n"); break;
  case GUMBO_NODE_WHITESPACE:
    break;
  }
}

static char *all_text(GumboNode *elt)
{
  buffer_t *B;
  GumboNode *C;
  int i;
  char *R;
  B= buffer_new(1024);
  for(i=0;i<elt->v.element.children.length;i++)
  {
    C= elt->v.element.children.data[i];
    if (C->type==GUMBO_NODE_TEXT || C->type==GUMBO_NODE_CDATA)
       buffer_write(B, (void*) C->v.text.text, strlen(C->v.text.text));
  }
  buffer_write(B, "", 1);
  buffer_free(B, (void**) &R, NULL);
  return R;
}

static GumboNode *find_first_dl(GumboNode *node)
{
  int i;
  GumboNode *C;
  GumboNode *R;
  GumboElement *elt;
  if (node->type!=GUMBO_NODE_ELEMENT) return NULL;
  elt= &node->v.element;
  for(i=0;i<elt->children.length;i++)
  {
    C= elt->children.data[i];
    if (C->type==GUMBO_NODE_ELEMENT && C->v.element.tag==GUMBO_TAG_DL)
      return C;
  }
  for(i=0;i<elt->children.length;i++)
  {
    C= elt->children.data[i];
    R= find_first_dl(C);
    if (R) return R;
  }
  return NULL;
}

static void dl_to_bkfolder(GumboNode *dlnode, bkelt_t *folder);
static void dl_item_to_bkelt(GumboNode *node, bkelt_t *folder)
{
  GumboElement *elt,*selt;
  GumboNode *snode;
  int i;
  GumboAttribute *attr;
  char *url,*date,*title;
  bkelt_t *B;
  int type;

  if (node->type!=GUMBO_NODE_ELEMENT) return ;
  elt= &node->v.element;
  if (elt->tag!=GUMBO_TAG_DT) return ;
  for(i=0;i<elt->children.length;i++)
  {
    snode= elt->children.data[i];
    if (snode->type!=GUMBO_NODE_ELEMENT) continue;
    selt= &snode->v.element;
    if (selt->tag==GUMBO_TAG_H3 || selt->tag==GUMBO_TAG_A) break;
  }
  if (i==elt->children.length) return ;

  attr= gumbo_get_attribute(&selt->attributes,"add_date");
  if (attr && attr->value) date= strdup(attr->value);
                      else date= strdup("");
  title= all_text(snode);

  if (selt->tag==GUMBO_TAG_A)
  {
    type= 'B';
    attr= gumbo_get_attribute(&selt->attributes,"href");
    if (attr && attr->value) url= strdup(attr->value);
                        else return ;
  }
  else
  {
    type= 'F';
    url= strdup("");
    for(i++;i<elt->children.length;i++)
    {
      snode= elt->children.data[i];
      if (snode->type!=GUMBO_NODE_ELEMENT) continue;
      selt= &snode->v.element;
      if (selt->tag==GUMBO_TAG_DL) break;
    }
    if (i==elt->children.length) snode= NULL;
  }

  B= bkelt(type, url,title,date);
  bkelt_add_child(folder,B);

  if (type=='F' && snode) 
    dl_to_bkfolder(snode, B);
}

static void dl_to_bkfolder(GumboNode *dlnode, bkelt_t *folder)
{
  GumboElement *elt;
  int i;
  if (dlnode->type!=GUMBO_NODE_ELEMENT) return ;
  elt= &dlnode->v.element;
  if (elt->tag!=GUMBO_TAG_DL) return ;
  for(i=0;i<elt->children.length;i++)
     dl_item_to_bkelt(elt->children.data[i], folder);
}

int html_to_bklist(char *fn, bkelt_t **Rtop)
{
  unsigned char *fd; size_t fl;
  bkelt_t *top;
  GumboNode *DL;
  if (file_read_all(fn, &fd, &fl))
    return 1;
  GumboOutput *gumbodoc;
  gumbodoc= gumbo_parse(fd);
  DL= find_first_dl(gumbodoc->root);
  // if (DL) prt_all(DL);
    //else  printf("Not found\n");
  top= bkelt('F',strdup(""),strdup(""),strdup(""));
  dl_to_bkfolder(DL, top);
  gumbo_destroy_output(&kGumboDefaultOptions,gumbodoc);
  free(fd);
  *Rtop= top;
  return 0;
}
